#include "phg.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef PHG_TO_P4EST
# include "phg-to-p4est.h"
#endif

int dim = Dim;

static void
func(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    if (dim == 1) {
	*values = 0.5 * (x * x + y * y + z * z);
	return;
    }
    *(values++) = 0.5 * x * x;
    *(values++) = 0.5 * y * y;
    if (dim >= 3) {
	*(values++) = 0.5 * z * z;
	if (dim > 3)
	    bzero(values, (dim - 3) * sizeof(*values));
    }
}

typedef enum {Mass = 0, Stiffness = 1} type_t;
static const char *type_name[] = {"mass", "stiffness"};

static MAT *
get_matrix(DOF *u, type_t type)
/* returns the mass matrix of the DOF "u" */
{
    MAP *map;
    MAT *A;
    GRID *g = u->g;
    ELEMENT *e = NULL;
    QCACHE *qc;
    int order, fid;

    phgDofSetDirichletBoundaryMask(u, 0);
    map = phgMapCreate(u, NULL);
    A = phgMapCreateMat(map, map);
    phgMapDestroy(&map);

    if (type == Mass) {
	order = 2 * DofTypeOrder(u, e);
	fid = Q_BAS;
    }
    else {
	order = 2 * DofTypeOrder(u, e) - 2;
	fid = Q_GRAD;
    }

    qc = phgQCNew(QD_PHG, u);
    ForAllElements(g, e) {
	int i, j, N = DofNBas(u, e);
	INT I, J, eno = e->index;
	FLOAT val, *rule;
	rule = phgQuadGetRule3D(g, e, order);
	phgQCSetRule(qc, rule, -1.);
	for (i = 0; i < N; i++) {
	    I = phgMapE2G(A->rmap, 0, e, i);
	    for (j = 0; j <= i; j++) {
		J = phgMapE2G(A->cmap, 0, e, j);
		val = phgQCIntegrate(qc, eno, fid, j, qc, eno, fid, i);
		phgMatAddGlobalEntry(A, I, J, val); 
		if (i != j)
		    phgMatAddGlobalEntry(A, J, I, val); 
	    }
	}
	phgFree(rule);
    }
    phgQCFree(&qc);
    phgMatAssemble(A);

    return A;
}

int
main(int argc, char **argv)
{
    GRID *g;
    DOF *u, *tmp;
    DOF_TYPE *vtype;
    char *fn = "../p4est/cube.mesh";
    INT refine0 = 0, refine = 0;
    BOOLEAN matlab = FALSE;
    int i, type;

    phgOptionsRegisterFilename("-mesh_file", "Mesh file", (char **)&fn);
    phgOptionsRegisterInt("-dim", "Vector dimension (1 or 3)", &dim);
    phgOptionsRegisterInt("-refine0", "Initial refinement levels", &refine0);
    phgOptionsRegisterInt("-refine", "Repeated refinement levels", &refine);
    phgOptionsRegisterNoArg("-matlab", "Dump MATLAB matrices", &matlab);

    //phgOptionsPreset("-dof_type P2 -refine 0 -mesh_file tetra.dat");
    phgInit(&argc, &argv);

    vtype = phgDofTypeVector(DOF_DEFAULT, dim, NULL);

    g = phgNewGrid(-1);
    phgImport(g, fn, TRUE);

    for (i = 0; i < refine0; i++) {
	phgRefineAllElements(g, 1);
	phgBalanceGrid(g, 1.2, 0, NULL, 1.0);
	phgPrintf("Initial refinement: %"dFMT" elements.\n", g->nelem_global);
    }
    

    u = phgDofNew(g, vtype, 1, "u", DofInterpolation);
    phgDofSetDataByFunction(u, func);
    //phgDofDump(u);
    tmp = phgDofGradient(u, NULL, NULL, "grad");
    phgDofSetFunction(tmp, DofInterpolation);
#if 0
    ELEMENT *e;
    assert(dim == Dim);
    ForAllElements(g, e) {
# ifdef PHG_TO_P4EST
	FLOAT xref[] = {0.5,0.5,0.5};
# else	/* defined(PHG_TO_P4EST) */
	FLOAT xref[] = {0.25,0.25,0.25,0.25};
# endif	/* defined(PHG_TO_P4EST) */
	FLOAT x[Dim], val[Dim * Dim];
	phgGeomLambda2XYZ(g, e, xref, x, x + 1, x + 2);
	phgDofEval(u, e, xref, val);
	phgInfo(-1, "u(%g,%g,%g) = (%g,%g,%g)\n",
			x[0], x[1], x[2], val[0], val[1], val[2]);
	phgDofEval(tmp, e, xref, val);
	phgInfo(-1, "grad(%g,%g,%g) = ((%g,%g,%g),(%g,%g,%g),(%g,%g,%g))\n",
			x[0], x[1], x[2], val[0], val[1], val[2],
					  val[3], val[4], val[5],
					  val[6], val[7], val[8]);
    }
#endif

    i = 0;
    while (i < refine) {
	i += 1;
	phgRefineAllElements(g, 1);
	phgBalanceGrid(g, 1.2, 0, NULL, 1.0);
	phgPrintf("Repeated refinement: %"dFMT" elements.\n", g->nelem_global);
    }
    phgPrintf("|u|_2 = %g, |grad(u)|_2 = %g\n",
		(double)phgDofNormL2(u), (double)phgDofNormL2(tmp));

    phgDofFree(&tmp);

    tmp = phgDofNew(g, vtype->base_type, 1, "u0", DofNoData);
    for (type = 0; type < sizeof(type_name) / sizeof(type_name[0]); type++) {
	MAT *A, *a;
	INT j;
	A = get_matrix(u, type);
	a = get_matrix(tmp, type);	/* matrix for the base type */
	if (matlab) {
	    char name0[strlen(type_name[type]) + 2];
	    sprintf(name0, "%s0", type_name[type]);
	    phgMatDumpMATLAB(A, type_name[type], NULL);
	    phgMatDumpMATLAB(a, name0, NULL);
	}
	/* check that "A" == P("a"), where P is the 1D->3D expansion operator */
	phgPrintf("Checking %s matrix: norm = %g ...",
		  type_name[type], (double)phgMatNormInfty(A));
	for (j = 0; j < a->rmap->nlocal; j++) {
	    const MAT_ROW *row = phgMatGetRow(a, j);
	    for (i = 0; i < row->ncols; i++) {
		/* Note: column indices in row are global */
		INT k = row->cols[i];
		FLOAT d0 = row->data[i];
		int ii;
		for (ii = 0; ii < dim; ii++) {
		    FLOAT d = phgMatGetLGEntry(A, j * dim + ii, k * dim + ii);
		    if (Fabs(d0 - d) > 100.0 * FLOAT_EPSILON * Fabs(d0)) {
			j += a->rmap->partition[g->rank]; /* -> global index */
			phgError(1, "check failed: A(%d,%d)=%g, a(%d,%d)=%g\n",
				    j * dim + ii, k * dim + ii, d, j, k, d0); 
		    }
		}
	    }
	}
	phgPrintf(" passed.\n");
	phgMatDestroy(&a);
	phgMatDestroy(&A);
    }
    phgDofFree(&tmp);

    phgDofFree(&u);

    phgDofTypeFree(&vtype);

    phgFreeGrid(&g);
    phgFinalize();

    return 0;
}
